from pystorm.utils.type_safety import ensure_numpy, ensure_torch
from scipy.signal import firwin
from math import ceil 
from pystorm import minitorch as mnt

def get_firwin(ntaps, centered_band, beta, fs):
    return firwin(ntaps, centered_band, window=("kaiser", beta), scale = True, pass_zero=False,fs=fs)

def get_fir_window(band, ripple, width, fs):
    """
    Shamelessly stolen from 
    https://github.com/scipy/scipy/blob/v0.14.0/scipy/signal/fir_filter_design.py#L85
    and 
    https://github.com/scipy/scipy/blob/v0.14.0/scipy/signal/fir_filter_design.py#L29 
    """
    width_norm = width/(0.5*fs)
    a = abs(ripple)  
    if a > 50:
        beta = 0.1102 * (a - 8.7)
    elif a > 21:
        beta = 0.5842 * (a - 21) ** 0.4 + 0.07886 * (a - 21)
    else:
        beta = 0.0
    beta = round(beta,4)
    numtaps = (a - 7.95) / 2.285 / (mnt.pi * width_norm) + 1
    ntaps = int(ceil(numtaps))
    ntaps = ntaps + (1-ntaps%2)
    centered_band = ensure_numpy([1e-5,fs//2 - 1])
    if band[0] is not None:
        centered_band[0] = band[0]-width/2
    if band[1] is not None:
        centered_band[1] = band[1]+width/2

    window = get_firwin(ntaps, centered_band,beta, fs)
    return window

def band_pass_torchaudio(signal,win, fs, return_pad = 0.2, convolve_type = "auto", device="cpu", verbose = 1, return_numba_compatible = False):
    signal = ensure_torch(signal)
    win = ensure_torch(win)
    
    if convolve_type == "fft" or (convolve_type == "auto" and len(win)*signal.shape[-1]>1e5 and abs(len(win)-signal.shape[-1])>1e2):
        convolve = mnt.tfftconvolve
    else:
        convolve = mnt.tconvolve

    win = win.to(device)

    E = win[win.shape[-1]//2:] ** 2
    E = E.cumsum(dim=-1)
    E = E / E.amax()
    iE99 = ((E-0.99).abs().argmin() / fs).item()
    edge_percent = 2*iE99 / (signal.shape[-1]/fs)
    if edge_percent>0.1:
        if verbose > 0:
            print(f"Start up and end transients represent {round(edge_percent*100,2)}% of your data.")

    pad_size = win.shape[0]//2
    mean_centered_signal = signal
    if len(mean_centered_signal.shape)<2:
        mean_centered_signal = mean_centered_signal.unsqueeze(0)
    signal_size = mean_centered_signal.shape[-1]
    mean_signal = mean_centered_signal.mean(-1, keepdims=True)
    mean_centered_signal = mean_centered_signal - mean_signal
    padded_signal = mnt.cat([mnt.zeros(*mean_centered_signal.shape[:-1],pad_size, device=device),mean_centered_signal,mnt.zeros(*mean_centered_signal.shape[:-1],pad_size, device=device)], dim = -1).to(device)
    if len(padded_signal.shape)<2:
        padded_signal = padded_signal.unsqueeze(0)
    if return_pad is not None:
        if return_pad > 1:
            return_pad = 1/return_pad
        signal_start = 2*int(pad_size)
        remainin_pad_size = int(pad_size*return_pad)
        remaining_margin_start = signal_start-remainin_pad_size
        remaining_margin_end = signal_start + signal_size + remainin_pad_size
    else:
        signal_start = 2*int(pad_size)
        remainin_pad_size = 2*int(pad_size)
        remaining_margin_start = None
        remaining_margin_end = None
    win = win.view(*[1 for _ in range(len(padded_signal.shape[:-1]))],-1)
    if win.dtype != padded_signal.dtype:
        print("correcting")
        win = ensure_torch(win,type_float=True)
        padded_signal = ensure_torch(padded_signal,type_float=True)
    filtered_signal = convolve(padded_signal,win, mode = "full")[...,remaining_margin_start:remaining_margin_end]
    signal_mask = mnt.zeros(filtered_signal.shape[-1], dtype=int)
    
    signal_end = signal_size+remainin_pad_size
    signal_indices = mnt.arange(remainin_pad_size,signal_end, dtype=int, device=device) 
    signal_mask[signal_indices] = 1
    if len(signal.shape)<2:
        filtered_signal = filtered_signal
    
    if return_numba_compatible: # For future use
        return_val = mnt.cat([filtered_signal,signal_mask.unsqueeze(0)], dim = 0).cpu()
        return ensure_numpy(return_val)
    else:
        return ensure_numpy(filtered_signal), ensure_numpy(signal_mask)

def band_pass(signal,fs,band,ripple = 60, width = 1.0, keep_pad_percent = 0.2,convolve_type = "auto",  return_with_pad = False, backend = "torch", device="cpu"):
    band = ensure_numpy(band)
    win = get_fir_window(band,ripple,width,fs)
    if backend == "torch":
        filtered_signal, filtered_mask = band_pass_torchaudio(signal,win,fs,return_pad=keep_pad_percent, convolve_type = convolve_type, device=device)
    else: # For future use
        raise NotImplementedError('The only backend available for now is "torch".')
    if return_with_pad:
        return filtered_signal, filtered_mask
    else:
        return filtered_signal[...,filtered_mask==1]



from pystorm.utils.type_safety import ensure_numpy, ensure_torch
from pystorm.signal_processing import band_pass
from pystorm import minitorch as mnt

def get_hilbert_torch(signal, fs, pad_size = None, window_mask = None, device = "cpu"):
    signal = ensure_torch(signal).to(device)
    
    if window_mask is not None: # Foreshadowing Amplitude Envelope Correlation implementation
        window_mask = ensure_torch(window_mask)
        windowed_signal = signal[..., window_mask]
    else:
        windowed_signal = signal
    
    if pad_size is not None:
        pad_size = int(pad_size*fs)
        padded_signal = mnt.cat([mnt.zeros(*signal.shape[:-1],pad_size, device=device),windowed_signal,mnt.zeros(*signal.shape[:-1],pad_size, device=device)], dim = -1).to(device)
    else:
        padded_signal = windowed_signal
        pad_size = 0
    
    recover_signal_mask = mnt.arange(windowed_signal.shape[-1])+pad_size
    freqs = mnt.fftfreq(padded_signal.shape[-1],d=1/fs)
    signal_fft = mnt.fft(padded_signal)
    signal_fft[...,freqs<0] = 0
    signal_fft[...,freqs>0] = signal_fft[...,freqs>0]*2
    analytical_signal = mnt.ifft(signal_fft)
    analytical_signal = analytical_signal[...,recover_signal_mask]

    return analytical_signal

def hilbert(signal, fs, pad_size = None, window_mask = None, return_envelope = False, backend="torch", device = "cpu"):
    if backend == "torch":
        analytical_signal = get_hilbert_torch(signal, fs, pad_size = pad_size, window_mask=window_mask, device=device)
    else: # For future use
        raise NotImplementedError('The only backend available for now is "torch".')
    if return_envelope:
        return ensure_numpy(analytical_signal.abs())
    return ensure_numpy(analytical_signal)


def band_pass_hilbert(signal,fs,band,ripple = 60, width = 1.0, keep_pad_percent_for_hilbert = 0.2, return_envelope = False, convolve_type = "auto", backend = "torch", device="cpu"):
    filtered_signal, signal_mask = band_pass(signal,fs,band,ripple=ripple,width=width,keep_pad_percent=keep_pad_percent_for_hilbert,convolve_type=convolve_type, return_with_pad=True, backend=backend,device=device)
    if backend == "torch":
        signal_mask = ensure_torch(signal_mask==1.0)
        pad_size = mnt.argwhere(signal_mask)[0]//fs
        analytical_signal = get_hilbert_torch(filtered_signal[...,signal_mask], fs, pad_size = pad_size, window_mask=None, device=device)
    else: # For future use
        raise NotImplementedError('The only backend available for now is "torch".')
    if return_envelope:
        return ensure_numpy(analytical_signal.abs())
    return ensure_numpy(analytical_signal)



"""
A module that serves as lightweight pytorch.

The intent is to possibly minimize RAM usage from functions that would be executed by multiple processes in parallel on the GPU.
"""

from torch.fft import rfft as trfft
from torch.fft import rfftfreq as trfftfreq 
from torch.fft import fft as tfft
from torch.fft import ifft as tifft
from torch.fft import fftfreq as tfftfreq 
from torch import cat as tcat
from torch import linspace as tlinspace
from torch import pi as tpi
from torch import arange as tarange
from torch import zeros as tzeros
from torch import zeros_like as  tzeros_like
from torch import ones_like as  tones_like
from torch import ones as tones
from torch import as_tensor as tas_tensor
from torch import from_numpy as tfrom_numpy
from torch import complex64, complex128
from torch import cos as tcos
from torch import eye as teye
from torch import argwhere as targwhere
from torchaudio.functional import convolve as tconvolve
from torchaudio.functional import fftconvolve as tfftconvolve
from torch import float16,float32,float64
from torch import set_default_dtype
pi = tpi

__default_dtype__ = float64
set_default_dtype(__default_dtype__)
def set_minitorch_default_dtype(default_type = "float64"):
    global __default_dtype__
    if default_type == "float16":
        __default_dtype__ = float16
        print("Warning: Pytorch only supports FFT of signals whose length are powers of 2 in this float type")
    elif default_type == "float32":
        __default_dtype__ = float32
    else:
        __default_dtype__ = float64
    set_default_dtype(__default_dtype__)
    

def eye(*args, **kwargs):
    """
    eye(n, m=None, *, out=None, dtype=None, layout=torch.strided, device=None, requires_grad=False) -> Tensor

    Returns a 2-D tensor with ones on the diagonal and zeros elsewhere.

    Args:
        n (int): the number of rows
        m (int, optional): the number of columns with default being :attr:`n`

    Keyword arguments:
        out (Tensor, optional): the output tensor.
        dtype (:class:`torch.dtype`, optional): the desired data type of returned tensor.
            Default: if ``None``, uses a global default (see :func:`torch.set_default_tensor_type`).
        layout (:class:`torch.layout`, optional): the desired layout of returned Tensor.
            Default: ``torch.strided``.
        device (:class:`torch.device`, optional): the desired device of returned tensor.
            Default: if ``None``, uses the current device for the default tensor type
            (see :func:`torch.set_default_tensor_type`). :attr:`device` will be the CPU
            for CPU tensor types and the current CUDA device for CUDA tensor types.
        requires_grad (bool, optional): If autograd should record operations on the
            returned tensor. Default: ``False``.

    Returns:
        Tensor: A 2-D tensor with ones on the diagonal and zeros elsewhere
        """
    return teye(*args,**kwargs)

def zeros(*args, **kwargs):
    """
    zeros(*size, *, out=None, dtype=None, layout=torch.strided, device=None, requires_grad=False) -> Tensor

    Returns a tensor filled with the scalar value `0`, with the shape defined
    by the variable argument :attr:`size`.

    Args:
        size (int...): a sequence of integers defining the shape of the output tensor.
            Can be a variable number of arguments or a collection like a list or tuple.

    Keyword args:
        out (Tensor, optional): the output tensor.
        dtype (:class:`torch.dtype`, optional): the desired data type of returned tensor.
            Default: if ``None``, uses a global default (see :func:`torch.set_default_tensor_type`).
        layout (:class:`torch.layout`, optional): the desired layout of returned Tensor.
            Default: ``torch.strided``.
        device (:class:`torch.device`, optional): the desired device of returned tensor.
            Default: if ``None``, uses the current device for the default tensor type
            (see :func:`torch.set_default_tensor_type`). :attr:`device` will be the CPU
            for CPU tensor types and the current CUDA device for CUDA tensor types.
        requires_grad (bool, optional): If autograd should record operations on the
            returned tensor. Default: ``False``.
            """
    return tzeros(*args,**kwargs)

def ones(*args, **kwargs):
    """
    ones(*size, *, out=None, dtype=None, layout=torch.strided, device=None, requires_grad=False) -> Tensor

    Returns a tensor filled with the scalar value `1`, with the shape defined
    by the variable argument :attr:`size`.

    Args:
        size (int...): a sequence of integers defining the shape of the output tensor.
            Can be a variable number of arguments or a collection like a list or tuple.

    Keyword arguments:
        out (Tensor, optional): the output tensor.
        dtype (:class:`torch.dtype`, optional): the desired data type of returned tensor.
            Default: if ``None``, uses a global default (see :func:`torch.set_default_tensor_type`).
        layout (:class:`torch.layout`, optional): the desired layout of returned Tensor.
            Default: ``torch.strided``.
        device (:class:`torch.device`, optional): the desired device of returned tensor.
            Default: if ``None``, uses the current device for the default tensor type
            (see :func:`torch.set_default_tensor_type`). :attr:`device` will be the CPU
            for CPU tensor types and the current CUDA device for CUDA tensor types.
        requires_grad (bool, optional): If autograd should record operations on the
            returned tensor. Default: ``False``.
        """
    return tones(*args,**kwargs)

def cat(*args, **kwargs):
    """
    cat(tensors, dim=0, *, out=None) -> Tensor

    Concatenates the given sequence of :attr:`seq` tensors in the given dimension.
    All tensors must either have the same shape (except in the concatenating
    dimension) or be empty.

    :func:`torch.cat` can be seen as an inverse operation for :func:`torch.split`
    and :func:`torch.chunk`.

    :func:`torch.cat` can be best understood via examples.

    .. seealso::

        :func:`torch.stack` concatenates the given sequence along a new dimension.

    Args:
        tensors (sequence of Tensors): any python sequence of tensors of the same type.
            Non-empty tensors provided must have the same shape, except in the
            cat dimension.
        dim (int, optional): the dimension over which the tensors are concatenated

    Keyword args:
        out (Tensor, optional): the output tensor.
    """
    return tcat(*args,**kwargs)

def arange(*args, **kwargs):
    """
    arange(start=0, end, step=1, *, out=None, dtype=None, layout=torch.strided, device=None, requires_grad=False) -> Tensor

    Returns a 1-D tensor of size :math:`\left\lceil \frac{\text{end} - \text{start}}{\text{step}} \right\rceil`
    with values from the interval ``[start, end)`` taken with common difference
    :attr:`step` beginning from `start`.

    Note that non-integer :attr:`step` is subject to floating point rounding errors when
    comparing against :attr:`end`; to avoid inconsistency, we advise subtracting a small epsilon from :attr:`end`
    in such cases.

    .. math::
        \text{out}_{{i+1}} = \text{out}_{i} + \text{step}

    Args:
        start (Number): the starting value for the set of points. Default: ``0``.
        end (Number): the ending value for the set of points
        step (Number): the gap between each pair of adjacent points. Default: ``1``.

    Keyword args:
        out (Tensor, optional): the output tensor.
        dtype (:class:`torch.dtype`, optional): the desired data type of returned tensor.
            Default: if ``None``, uses a global default (see :func:`torch.set_default_tensor_type`). If `dtype` is not given, infer the data type from the other input
            arguments. If any of `start`, `end`, or `stop` are floating-point, the
            `dtype` is inferred to be the default dtype, see
            :meth:`~torch.get_default_dtype`. Otherwise, the `dtype` is inferred to
            be `torch.int64`.
        layout (:class:`torch.layout`, optional): the desired layout of returned Tensor.
            Default: ``torch.strided``.
        device (:class:`torch.device`, optional): the desired device of returned tensor.
            Default: if ``None``, uses the current device for the default tensor type
            (see :func:`torch.set_default_tensor_type`). :attr:`device` will be the CPU
            for CPU tensor types and the current CUDA device for CUDA tensor types.
        requires_grad (bool, optional): If autograd should record operations on the
            returned tensor. Default: ``False``.

    """
    return tarange(*args,**kwargs)

def linspace(*args, **kwargs):
    """
    linspace(start, end, steps, *, out=None, dtype=None, layout=torch.strided, device=None, requires_grad=False) -> Tensor

    Creates a one-dimensional tensor of size :attr:`steps` whose values are evenly
    spaced from :attr:`start` to :attr:`end`, inclusive. That is, the value are:

    .. math::
        (\text{start},
        \text{start} + \frac{\text{end} - \text{start}}{\text{steps} - 1},
        \ldots,
        \text{start} + (\text{steps} - 2) * \frac{\text{end} - \text{start}}{\text{steps} - 1},
        \text{end})


    From PyTorch 1.11 linspace requires the steps argument. Use steps=100 to restore the previous behavior.

    Args:
        start (float or Tensor): the starting value for the set of points. If `Tensor`, it must be 0-dimensional
        end (float or Tensor): the ending value for the set of points. If `Tensor`, it must be 0-dimensional
        steps (int): size of the constructed tensor

    Keyword arguments:
        out (Tensor, optional): the output tensor.
        dtype (torch.dtype, optional): the data type to perform the computation in.
            Default: if None, uses the global default dtype (see torch.get_default_dtype())
            when both :attr:`start` and :attr:`end` are real,
            and corresponding complex dtype when either is complex.
        layout (:class:`torch.layout`, optional): the desired layout of returned Tensor.
            Default: ``torch.strided``.
        device (:class:`torch.device`, optional): the desired device of returned tensor.
            Default: if ``None``, uses the current device for the default tensor type
            (see :func:`torch.set_default_tensor_type`). :attr:`device` will be the CPU
            for CPU tensor types and the current CUDA device for CUDA tensor types.
        requires_grad (bool, optional): If autograd should record operations on the
            returned tensor. Default: ``False``.
    """
    return tlinspace(*args,**kwargs)

def ones_like(*args, **kwargs):
    """
    ones_like(input, *, dtype=None, layout=None, device=None, requires_grad=False, memory_format=torch.preserve_format) -> Tensor

    Returns a tensor filled with the scalar value `1`, with the same size as
    :attr:`input`. ``torch.ones_like(input)`` is equivalent to
    ``torch.ones(input.size(), dtype=input.dtype, layout=input.layout, device=input.device)``.

    .. warning::
        As of 0.4, this function does not support an :attr:`out` keyword. As an alternative,
        the old ``torch.ones_like(input, out=output)`` is equivalent to
        ``torch.ones(input.size(), out=output)``.

    Args:
        input (Tensor): the size of :attr:`input` will determine size of the output tensor.

    Keyword arguments:
        dtype (:class:`torch.dtype`, optional): the desired data type of returned Tensor.
            Default: if ``None``, defaults to the dtype of :attr:`input`.
        layout (:class:`torch.layout`, optional): the desired layout of returned tensor.
            Default: if ``None``, defaults to the layout of :attr:`input`.
        device (:class:`torch.device`, optional): the desired device of returned tensor.
            Default: if ``None``, defaults to the device of :attr:`input`.
        requires_grad (bool, optional): If autograd should record operations on the
            returned tensor. Default: ``False``.
        memory_format (:class:`torch.memory_format`, optional): the desired memory format of
            returned Tensor. Default: ``torch.preserve_format``.

    """
    return tones_like(*args,**kwargs)

def zeros_like(*args, **kwargs):
    """
    zeros_like(input, *, dtype=None, layout=None, device=None, requires_grad=False, memory_format=torch.preserve_format) -> Tensor

    Returns a tensor filled with the scalar value `0`, with the same size as
    :attr:`input`. ``torch.zeros_like(input)`` is equivalent to
    ``torch.zeros(input.size(), dtype=input.dtype, layout=input.layout, device=input.device)``.

    .. warning::
        As of 0.4, this function does not support an :attr:`out` keyword. As an alternative,
        the old ``torch.zeros_like(input, out=output)`` is equivalent to
        ``torch.zeros(input.size(), out=output)``.

    Args:
        input (Tensor): the size of :attr:`input` will determine size of the output tensor.

    Keyword args:
        dtype (:class:`torch.dtype`, optional): the desired data type of returned Tensor.
            Default: if ``None``, defaults to the dtype of :attr:`input`.
        layout (:class:`torch.layout`, optional): the desired layout of returned tensor.
            Default: if ``None``, defaults to the layout of :attr:`input`.
        device (:class:`torch.device`, optional): the desired device of returned tensor.
            Default: if ``None``, defaults to the device of :attr:`input`.
        requires_grad (bool, optional): If autograd should record operations on the
            returned tensor. Default: ``False``.
        memory_format (:class:`torch.memory_format`, optional): the desired memory format of
            returned Tensor. Default: ``torch.preserve_format``.

    """
    return tzeros_like(*args,**kwargs)

def as_tensor(*args,**kwargs):
    """
    as_tensor(data, dtype=None, device=None) -> Tensor

    Converts :attr:`data` into a tensor, sharing data and preserving autograd
    history if possible.

    If :attr:`data` is already a tensor with the requested dtype and device
    then :attr:`data` itself is returned, but if :attr:`data` is a
    tensor with a different dtype or device then it's copied as if using
    `data.to(dtype=dtype, device=device)`.

    If :attr:`data` is a NumPy array (an ndarray) with the same dtype and device then a
    tensor is constructed using :func:`torch.from_numpy`.

    .. seealso::

        :func:`torch.tensor` never shares its data and creates a new "leaf tensor" (see :doc:`/notes/autograd`).


    Args:
        data (array_like): Initial data for the tensor. Can be a list, tuple,
            NumPy ``ndarray``, scalar, and other types.
        dtype (:class:`torch.dtype`, optional): the desired data type of returned tensor.
            Default: if ``None``, infers data type from :attr:`data`.
        device (:class:`torch.device`, optional): the device of the constructed tensor. If None and data is a tensor
            then the device of data is used. If None and data is not a tensor then
            the result tensor is constructed on the current device.

    """
    return tas_tensor(*args,**kwargs)


def from_numpy(*args,**kwargs):
    """
        
    from_numpy(ndarray) -> Tensor

    Creates a :class:`Tensor` from a :class:`numpy.ndarray`.

    The returned tensor and :attr:`ndarray` share the same memory. Modifications to
    the tensor will be reflected in the :attr:`ndarray` and vice versa. The returned
    tensor is not resizable.

    It currently accepts :attr:`ndarray` with dtypes of ``numpy.float64``,
    ``numpy.float32``, ``numpy.float16``, ``numpy.complex64``, ``numpy.complex128``,
    ``numpy.int64``, ``numpy.int32``, ``numpy.int16``, ``numpy.int8``, ``numpy.uint8``,
    and ``bool``.

    .. warning::
        Writing to a tensor created from a read-only NumPy array is not supported and will result in undefined behavior.

    """
    return tfrom_numpy(*args,**kwargs)

def rfft(*args,**kwargs):
    """
    rfft(input, n=None, dim=-1, norm=None, *, out=None) -> Tensor

    Computes the one dimensional Fourier transform of real-valued :attr:`input`.

    The FFT of a real signal is Hermitian-symmetric, ``X[i] = conj(X[-i])`` so
    the output contains only the positive frequencies below the Nyquist frequency.
    To compute the full output, use :func:`~torch.fft.fft`

    Note:
        Supports torch.half on CUDA with GPU Architecture SM53 or greater.
        However it only supports powers of 2 signal length in every transformed dimension.

    Args:
        input (Tensor): the real input tensor
        n (int, optional): Signal length. If given, the input will either be zero-padded
            or trimmed to this length before computing the real FFT.
        dim (int, optional): The dimension along which to take the one dimensional real FFT.
        norm (str, optional): Normalization mode. For the forward transform
            (:func:`~torch.fft.rfft`), these correspond to:

            * ``"forward"`` - normalize by ``1/n``
            * ``"backward"`` - no normalization
            * ``"ortho"`` - normalize by ``1/sqrt(n)`` (making the FFT orthonormal)

            Calling the backward transform (:func:`~torch.fft.irfft`) with the same
            normalization mode will apply an overall normalization of ``1/n`` between
            the two transforms. This is required to make :func:`~torch.fft.irfft`
            the exact inverse.

            Default is ``"backward"`` (no normalization).

    Keyword args:
        out (Tensor, optional): the output tensor.
    """
    return trfft(*args,**kwargs)
    
def rfftfreq(*args,**kwargs):
    """
    rfftfreq(n, d=1.0, *, out=None, dtype=None, layout=torch.strided, device=None, requires_grad=False) -> Tensor

    Computes the sample frequencies for :func:`~torch.fft.rfft` with a signal of size :attr:`n`.

    Note:
        :func:`~torch.fft.rfft` returns Hermitian one-sided output, so only the
        positive frequency terms are returned. For a real FFT of length :attr:`n`
        and with inputs spaced in length unit :attr:`d`, the frequencies are::

            f = torch.arange((n + 1) // 2) / (d * n)

    Note:
        For even lengths, the Nyquist frequency at ``f[n/2]`` can be thought of as
        either negative or positive. Unlike :func:`~torch.fft.fftfreq`,
        :func:`~torch.fft.rfftfreq` always returns it as positive.

    Args:
        n (int): the real FFT length
        d (float, optional): The sampling length scale.
            The spacing between individual samples of the FFT input.
            The default assumes unit spacing, dividing that result by the actual
            spacing gives the result in physical frequency units.

    Keyword Args:
        out (Tensor, optional): the output tensor.
        dtype (:class:`torch.dtype`, optional): the desired data type of returned tensor.
            Default: if ``None``, uses a global default (see :func:`torch.set_default_tensor_type`).
        layout (:class:`torch.layout`, optional): the desired layout of returned Tensor.
            Default: ``torch.strided``.
        device (:class:`torch.device`, optional): the desired device of returned tensor.
            Default: if ``None``, uses the current device for the default tensor type
            (see :func:`torch.set_default_tensor_type`). :attr:`device` will be the CPU
            for CPU tensor types and the current CUDA device for CUDA tensor types.
        requires_grad (bool, optional): If autograd should record operations on the
            returned tensor. Default: ``False``.
    """
    return trfftfreq(*args,**kwargs)

def cos(*args,**kwargs):
    """
        
    cos(input, *, out=None) -> Tensor

    Returns a new tensor with the cosine  of the elements of :attr:`input`.

    .. math::
        \text{out}_{i} = \cos(\text{input}_{i})

    Args:
        input (Tensor): the input tensor.

    Keyword args:
        out (Tensor, optional): the output tensor.
    """
    return tcos(*args,**kwargs)


def fft(*args,**kwargs):
    """  
    fft(input, n=None, dim=-1, norm=None, *, out=None) -> Tensor

    Computes the one dimensional discrete Fourier transform of :attr:`input`.

    Note:
        The Fourier domain representation of any real signal satisfies the
        Hermitian property: `X[i] = conj(X[-i])`. This function always returns both
        the positive and negative frequency terms even though, for real inputs, the
        negative frequencies are redundant. :func:`~torch.fft.rfft` returns the
        more compact one-sided representation where only the positive frequencies
        are returned.

    Note:
        Supports torch.half and torch.chalf on CUDA with GPU Architecture SM53 or greater.
        However it only supports powers of 2 signal length in every transformed dimension.

    Args:
        input (Tensor): the input tensor
        n (int, optional): Signal length. If given, the input will either be zero-padded
            or trimmed to this length before computing the FFT.
        dim (int, optional): The dimension along which to take the one dimensional FFT.
        norm (str, optional): Normalization mode. For the forward transform
            (:func:`~torch.fft.fft`), these correspond to:

            * ``"forward"`` - normalize by ``1/n``
            * ``"backward"`` - no normalization
            * ``"ortho"`` - normalize by ``1/sqrt(n)`` (making the FFT orthonormal)

            Calling the backward transform (:func:`~torch.fft.ifft`) with the same
            normalization mode will apply an overall normalization of ``1/n`` between
            the two transforms. This is required to make :func:`~torch.fft.ifft`
            the exact inverse.

            Default is ``"backward"`` (no normalization).

    Keyword args:
        out (Tensor, optional): the output tensor.
    """
    return tfft(*args,**kwargs)

def ifft(*args,**kwargs):
    """
    ifft(input, n=None, dim=-1, norm=None, *, out=None) -> Tensor

    Computes the one dimensional inverse discrete Fourier transform of :attr:`input`.

    Note:
        Supports torch.half and torch.chalf on CUDA with GPU Architecture SM53 or greater.
        However it only supports powers of 2 signal length in every transformed dimension.

    Args:
        input (Tensor): the input tensor
        n (int, optional): Signal length. If given, the input will either be zero-padded
            or trimmed to this length before computing the IFFT.
        dim (int, optional): The dimension along which to take the one dimensional IFFT.
        norm (str, optional): Normalization mode. For the backward transform
            (:func:`~torch.fft.ifft`), these correspond to:

            * ``"forward"`` - no normalization
            * ``"backward"`` - normalize by ``1/n``
            * ``"ortho"`` - normalize by ``1/sqrt(n)`` (making the IFFT orthonormal)

            Calling the forward transform (:func:`~torch.fft.fft`) with the same
            normalization mode will apply an overall normalization of ``1/n`` between
            the two transforms. This is required to make :func:`~torch.fft.ifft`
            the exact inverse.

            Default is ``"backward"`` (normalize by ``1/n``).

    Keyword args:
        out (Tensor, optional): the output tensor.
    """
    return tifft(*args,**kwargs)

def fftfreq(*args,**kwargs):
    """
    fftfreq(n, d=1.0, *, out=None, dtype=None, layout=torch.strided, device=None, requires_grad=False) -> Tensor

    Computes the discrete Fourier Transform sample frequencies for a signal of size :attr:`n`.

    Note:
        By convention, :func:`~torch.fft.fft` returns positive frequency terms
        first, followed by the negative frequencies in reverse order, so that
        ``f[-i]`` for all :math:`0 < i \leq n/2`` in Python gives the negative
        frequency terms. For an FFT of length :attr:`n` and with inputs spaced in
        length unit :attr:`d`, the frequencies are::

            f = [0, 1, ..., (n - 1) // 2, -(n // 2), ..., -1] / (d * n)

    Note:
        For even lengths, the Nyquist frequency at ``f[n/2]`` can be thought of as
        either negative or positive. :func:`~torch.fft.fftfreq` follows NumPy's
        convention of taking it to be negative.

    Args:
        n (int): the FFT length
        d (float, optional): The sampling length scale.
            The spacing between individual samples of the FFT input.
            The default assumes unit spacing, dividing that result by the actual
            spacing gives the result in physical frequency units.

    Keyword Args:
        out (Tensor, optional): the output tensor.
        dtype (:class:`torch.dtype`, optional): the desired data type of returned tensor.
            Default: if ``None``, uses a global default (see :func:`torch.set_default_dtype`).
        layout (:class:`torch.layout`, optional): the desired layout of returned Tensor.
            Default: ``torch.strided``.
        device (:class:`torch.device`, optional): the desired device of returned tensor.
            Default: if ``None``, uses the current device for the default tensor type
            (see :func:`torch.set_default_device`). :attr:`device` will be the CPU
            for CPU tensor types and the current CUDA device for CUDA tensor types.
        requires_grad (bool, optional): If autograd should record operations on the
            returned tensor. Default: ``False``.

    """
    return tfftfreq(*args,**kwargs)


def convolve(*args, **kwargs):
    r"""
    Convolves inputs along their last dimension using the direct method.
    Note that, in contrast to :meth:`torch.nn.functional.conv1d`, which actually applies the valid cross-correlation
    operator, this function applies the true `convolution`_ operator.

    .. devices:: CPU CUDA

    .. properties:: Autograd TorchScript

    Args:
        x (torch.Tensor): First convolution operand, with shape `(..., N)`.
        y (torch.Tensor): Second convolution operand, with shape `(..., M)`
            (leading dimensions must be broadcast-able with those of ``x``).
        mode (str, optional): Must be one of ("full", "valid", "same").

            * "full": Returns the full convolution result, with shape `(..., N + M - 1)`. (Default)
            * "valid": Returns the segment of the full convolution result corresponding to where
              the two inputs overlap completely, with shape `(..., max(N, M) - min(N, M) + 1)`.
            * "same": Returns the center segment of the full convolution result, with shape `(..., N)`.

    Returns:
        torch.Tensor: Result of convolving ``x`` and ``y``, with shape `(..., L)`, where
        the leading dimensions match those of ``x`` and `L` is dictated by ``mode``.

    .. _convolution:
        https://en.wikipedia.org/wiki/Convolution
    """
    return tconvolve(*args,**kwargs)

    
def fftconvolve(*args, **kwargs):
    r"""
    Convolves inputs along their last dimension using FFT. For inputs with large last dimensions, this function
    is generally much faster than :meth:`convolve`.
    Note that, in contrast to :meth:`torch.nn.functional.conv1d`, which actually applies the valid cross-correlation
    operator, this function applies the true `convolution`_ operator.
    Also note that this function can only output float tensors (int tensor inputs will be cast to float).

    .. devices:: CPU CUDA

    .. properties:: Autograd TorchScript

    Args:
        x (torch.Tensor): First convolution operand, with shape `(..., N)`.
        y (torch.Tensor): Second convolution operand, with shape `(..., M)`
            (leading dimensions must be broadcast-able with those of ``x``).
        mode (str, optional): Must be one of ("full", "valid", "same").

            * "full": Returns the full convolution result, with shape `(..., N + M - 1)`. (Default)
            * "valid": Returns the segment of the full convolution result corresponding to where
              the two inputs overlap completely, with shape `(..., max(N, M) - min(N, M) + 1)`.
            * "same": Returns the center segment of the full convolution result, with shape `(..., N)`.

    Returns:
        torch.Tensor: Result of convolving ``x`` and ``y``, with shape `(..., L)`, where
        the leading dimensions match those of ``x`` and `L` is dictated by ``mode``.

    .. _convolution:
        https://en.wikipedia.org/wiki/Convolution
    """
    return tfftconvolve(*args,**kwargs)


def argwhere(*args, **kwargs):
    """
    argwhere(input) -> Tensor

    Returns a tensor containing the indices of all non-zero elements of
    :attr:`input`.  Each row in the result contains the indices of a non-zero
    element in :attr:`input`. The result is sorted lexicographically, with
    the last index changing the fastest (C-style).

    If :attr:`input` has :math:`n` dimensions, then the resulting indices tensor
    :attr:`out` is of size :math:`(z \times n)`, where :math:`z` is the total number of
    non-zero elements in the :attr:`input` tensor.

    .. note::
        This function is similar to NumPy's `argwhere`.

        When :attr:`input` is on CUDA, this function causes host-device synchronization.

    Args:
        {input}

    """
    return targwhere(*args,**kwargs)



from pystorm.utils import minitorch as mnt
from pystorm.utils.type_safety import ensure_torch, ensure_numpy

def welch_psd_source_space(time_series, fs, ker = None,window_length=2000, overlap = 0.5, device = "cpu"):
    if ker is None:
        return welch_psd_sensor_space(time_series, fs, window_length=window_length, overlap = overlap, device = device)
    nTime = time_series.shape[-1]
    Lwin = round(window_length/1000*fs)
    Loverlap = round(Lwin * overlap)
    Lwin = Lwin - Lwin%2
    Nwin = int((nTime - Loverlap)/(Lwin-Loverlap))

    freqs = mnt.rfftfreq(Lwin,1/fs).to(device)
    ker_cuda = ensure_torch(ker).to(device).type(mnt.complex64)
    sensor_cuda = ensure_torch(time_series).to(device)
    hamming_window = (0.54 - 0.46 * mnt.cos(mnt.linspace(0,2*mnt.pi,Lwin))).to(device)

    win_noise_power_gain = (hamming_window**2).sum()
    scaling_term = (2/(win_noise_power_gain*fs)).sqrt()
    time_sequence = mnt.arange(0,Lwin, dtype = int)

    fft_blocks = 0
    for i in range(Nwin):
        iTimes =  time_sequence + (i - 1)*(Lwin-Loverlap)
        block = sensor_cuda[:,iTimes]
        source_ts = (block) * hamming_window
        fft_res = mnt.rfft(source_ts,dim = -1) * scaling_term
        fft_res[:,1:] /= 2**(1/2)
        fft_blocks += (ker_cuda @ fft_res).abs()**2/Nwin
    return ensure_numpy(fft_blocks), ensure_numpy(freqs)

def welch_psd_sensor_space(time_series, fs, window_length=2000, overlap = 0.5, device = "cpu"):
    nTime = time_series.shape[-1]
    Lwin = round(window_length/1000*fs)
    Loverlap = round(Lwin * overlap)
    Lwin = Lwin - Lwin%2
    Nwin = int((nTime - Loverlap)/(Lwin-Loverlap))

    freqs = mnt.rfftfreq(Lwin,1/fs).to(device)
    sensor_cuda = ensure_torch(time_series).to(device)
    hamming_window = (0.54 - 0.46 * mnt.cos(mnt.linspace(0,2*mnt.pi,Lwin))).to(device)

    win_noise_power_gain = (hamming_window**2).sum()
    scaling_term = (2/(win_noise_power_gain*fs)).sqrt()
    time_sequence = mnt.arange(0,Lwin, dtype = int)

    fft_blocks = 0
    for i in range(Nwin):
        iTimes =  time_sequence + (i - 1)*(Lwin-Loverlap)
        block = sensor_cuda[:,iTimes]
        source_ts = (block) * hamming_window
        fft_res = mnt.rfft(source_ts,dim = -1) * scaling_term
        fft_res[:,1:] /= 2**(1/2)
        fft_blocks += (fft_res).abs()**2/Nwin
    return ensure_numpy(fft_blocks), ensure_numpy(freqs)


from .minitorch import as_tensor, from_numpy, __default_dtype__

def ensure_torch(x, type_float = False):
    try:
        x = as_tensor(x)
        if type_float:
            x = x.type(__default_dtype__)
    except:
        try:
            x = from_numpy(x)
            if type_float:
                x = x.type(__default_dtype__)
        except:
            pass
    
    if type_float:
        try:
            x = x.type(__default_dtype__)
        except:
            pass
    return x

def ensure_numpy(x):
    
    try:
        x = x.detach()
    except:
        pass
    
    try:
        x = x.to('cpu')
    except:
        pass
    
    try:
        x = x.numpy()
    except:
        pass
    
    return x


from setuptools import setup, find_packages
from os.path import join as os_join

# Get the current version number from inside the module
with open(os_join('pystorm', 'version.py')) as version_file:
    exec(version_file.read())

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
   name='pystorm3',
   version=__version__,
   description='Python implementation of some Brainstorm functions',
   license="GPL-3.0 license",
   long_description=long_description,
   long_description_content_type = 'text/markdown',
   author='Dominic Boutet',
   author_email='dominic.boutet@mail.mcgill.ca',
   maintainer='Dominic Boutet',
   maintainer_email='dominic.boutet@mail.mcgill.ca',
   url='https://github.com/NeuroLife77/pystorm',
   packages=find_packages(),
   install_requires=[
                        'torch>=2.2.0',
                        'numpy>=1.26',
                        'scipy>=1.13.0',
                        'torchaudio>=2.2.0'
    ],
   python_requires='>=3.9',
   keywords = [
                    'neuroscience',
                    'neuroimaging',
                    'neural oscillations',
                    'time series analysis',
                    'time frequency analysis',
                    'electrophysiology'
    ],
   classifiers=[
                'Development Status :: 1 - Planning',
                'Intended Audience :: Science/Research',
                'Topic :: Scientific/Engineering',
                'Operating System :: Unix',
                'Programming Language :: Python',
   ]
   #scripts=[
   #        ]
)